@@ -1,35 +1,48 @@
+from __future__ import print_function
+
 import argparse
+import itertools
 import os
 import random
-from fnmatch import fnmatchcase
-import sys
 import re
 import shlex
 import string
-import pkg_resources
-import itertools
-import pluggy
+import sys
+from fnmatch import fnmatchcase
 from subprocess import list2cmdline
 
-import tox.interpreters
-from tox import hookspecs
-from tox._verlib import NormalizedVersion
-
+import pkg_resources
+import pluggy
 import py
 
 import tox
+import tox.interpreters
+from tox import hookspecs
+from tox._verlib import NormalizedVersion
 
 iswin32 = sys.platform == "win32"
 
 default_factors = {'jython': 'jython', 'pypy': 'pypy', 'pypy3': 'pypy3',
                    'py': sys.executable, 'py2': 'python2', 'py3': 'python3'}
-for version in '26,27,32,33,34,35,36,37'.split(','):
+for version in '27,34,35,36,37'.split(','):
     default_factors['py' + version] = 'python%s.%s' % tuple(version)
 
 hookimpl = pluggy.HookimplMarker("tox")
 
 _dummy = object()
 
+PIP_INSTALL_SHORT_OPTIONS_ARGUMENT = ['-{}'.format(option) for option in [
+    'c', 'e', 'r', 'b', 't', 'd',
+]]
+
+PIP_INSTALL_LONG_OPTIONS_ARGUMENT = ['--{}'.format(option) for option in [
+    'constraint', 'editable', 'requirement', 'build', 'target', 'download',
+    'src', 'upgrade-strategy', 'install-options', 'global-option',
+    'root', 'prefix', 'no-binary', 'only-binary', 'index-url',
+    'extra-index-url', 'find-links', 'proxy', 'retries', 'timeout',
+    'exists-action', 'trusted-host', 'client-cert', 'cache-dir',
+]]
+
 
 def get_plugin_manager(plugins=()):
     # initialize plugin manager
@@ -123,6 +136,24 @@ class DepOption:
             else:
                 name = depline.strip()
                 ixserver = None
+
+                # we need to process options, in case they contain a space,
+                # as the subprocess call to pip install will otherwise fail.
+
+                # in case of a short option, we remove the space
+                for option in PIP_INSTALL_SHORT_OPTIONS_ARGUMENT:
+                    if name.startswith(option):
+                        name = '{}{}'.format(
+                            option, name[len(option):].strip()
+                        )
+
+                # in case of a long option, we add an equal sign
+                for option in PIP_INSTALL_LONG_OPTIONS_ARGUMENT:
+                    if name.startswith(option + ' '):
+                        name = '{}={}'.format(
+                            option, name[len(option):].strip()
+                        )
+
             name = self._replace_forced_dep(name, config)
             deps.append(DepConfig(name, ixserver))
         return deps
@@ -193,19 +224,15 @@ class InstallcmdOption:
         return value
 
 
-def parseconfig(args=None, plugins=()):
+def parseconfig(args, plugins=()):
     """
-    :param list[str] args: Optional list of arguments.
+    :param list[str] args: list of arguments.
     :type pkg: str
     :rtype: :class:`Config`
     :raise SystemExit: toxinit file is not found
     """
 
     pm = get_plugin_manager(plugins)
-
-    if args is None:
-        args = sys.argv[1:]
-
     # prepare command line options
     parser = Parser()
     pm.hook.tox_addoption(parser=parser)
@@ -216,7 +243,9 @@ def parseconfig(args=None, plugins=()):
     config = Config(pluginmanager=pm, option=option, interpreters=interpreters)
     config._parser = parser
     config._testenv_attr = parser._testenv_attr
-
+    if config.option.version:
+        print(get_version_info(pm))
+        raise SystemExit(0)
     # parse ini file
     basename = config.option.configfile
     if os.path.isfile(basename):
@@ -243,7 +272,7 @@ def parseconfig(args=None, plugins=()):
     except tox.exception.InterpreterNotFound:
         exn = sys.exc_info()[1]
         # Use stdout to match test expectations
-        py.builtin.print_("ERROR: " + str(exn))
+        print("ERROR: " + str(exn))
 
     # post process config object
     pm.hook.tox_configure(config=config)
@@ -252,25 +281,33 @@ def parseconfig(args=None, plugins=()):
 
 
 def feedback(msg, sysexit=False):
-    py.builtin.print_("ERROR: " + msg, file=sys.stderr)
+    print("ERROR: " + msg, file=sys.stderr)
     if sysexit:
         raise SystemExit(1)
 
 
-class VersionAction(argparse.Action):
-    def __call__(self, argparser, *args, **kwargs):
-        version = tox.__version__
-        py.builtin.print_("%s imported from %s" % (version, tox.__file__))
-        raise SystemExit(0)
+def get_version_info(pm):
+    out = ["%s imported from %s" % (tox.__version__, tox.__file__)]
+    plugin_dist_info = pm.list_plugin_distinfo()
+    if plugin_dist_info:
+        out.append('registered plugins:')
+        for mod, egg_info in plugin_dist_info:
+            source = getattr(mod, '__file__', repr(mod))
+            out.append("    %s-%s at %s" % (
+                egg_info.project_name, egg_info.version, source))
+    return '\n'.join(out)
 
 
-class SetenvDict:
-    def __init__(self, dict, reader):
+class SetenvDict(object):
+    def __init__(self, definitions, reader):
+        self.definitions = definitions
         self.reader = reader
-        self.definitions = dict
         self.resolved = {}
         self._lookupstack = []
 
+    def __repr__(self):
+        return "%s: %s" % (self.__class__.__name__, self.definitions)
+
     def __contains__(self, name):
         return name in self.definitions
 
@@ -308,16 +345,17 @@ class SetenvDict:
 @hookimpl
 def tox_addoption(parser):
     # formatter_class=argparse.ArgumentDefaultsHelpFormatter)
-    parser.add_argument("--version", nargs=0, action=VersionAction,
-                        dest="version",
+    parser.add_argument("--version", action="store_true", dest="version",
                         help="report version information to stdout.")
     parser.add_argument("-h", "--help", action="store_true", dest="help",
                         help="show help about options")
     parser.add_argument("--help-ini", "--hi", action="store_true", dest="helpini",
                         help="show help about ini-names")
-    parser.add_argument("-v", action='count', dest="verbosity", default=0,
+    parser.add_argument("-v", action='count', dest="verbose_level", default=0,
                         help="increase verbosity of reporting output. -vv mode turns off "
-                        "output redirection for package installation")
+                             "output redirection for package installation")
+    parser.add_argument("-q", action="count", dest="quiet_level", default=0,
+                        help="progressively silence reporting output.")
     parser.add_argument("--showconfig", action="store_true",
                         help="show configuration information for all environments. ")
     parser.add_argument("-l", "--listenvs", action="store_true",
@@ -358,7 +396,7 @@ def tox_addoption(parser):
     parser.add_argument("--result-json", action="store",
                         dest="resultjson", metavar="PATH",
                         help="write a json file with detailed information "
-                        "about all commands and results involved.")
+                             "about all commands and results involved.")
 
     # We choose 1 to 4294967295 because it is the range of PYTHONHASHSEED.
     parser.add_argument("--hashseed", action="store",
@@ -468,9 +506,9 @@ def tox_addoption(parser):
             itertools.chain.from_iterable(
                 [x.split(' ') for x in value]))
 
-        passenv = set([
+        passenv = {
             "PATH", "PIP_INDEX_URL", "LANG", "LANGUAGE", "LD_LIBRARY_PATH"
-        ])
+        }
 
         # read in global passenv settings
         p = os.environ.get("TOX_TESTENV_PASSENV", None)
@@ -484,16 +522,17 @@ def tox_addoption(parser):
         # so we just pass it on by default for now.
         if sys.platform == "win32":
             passenv.add("SYSTEMDRIVE")  # needed for pip6
-            passenv.add("SYSTEMROOT")   # needed for python's crypto module
-            passenv.add("PATHEXT")      # needed for discovering executables
-            passenv.add("COMSPEC")      # needed for distutils cygwincompiler
+            passenv.add("SYSTEMROOT")  # needed for python's crypto module
+            passenv.add("PATHEXT")  # needed for discovering executables
+            passenv.add("COMSPEC")  # needed for distutils cygwincompiler
             passenv.add("TEMP")
             passenv.add("TMP")
             # for `multiprocessing.cpu_count()` on Windows
             # (prior to Python 3.4).
             passenv.add("NUMBER_OF_PROCESSORS")
+            passenv.add("PROCESSOR_ARCHITECTURE")  # platform.machine()
             passenv.add("USERPROFILE")  # needed for `os.path.expanduser()`
-            passenv.add("MSYSTEM")      # fixes #429
+            passenv.add("MSYSTEM")  # fixes #429
         else:
             passenv.add("TMPDIR")
         for spec in value:
@@ -579,6 +618,7 @@ def tox_addoption(parser):
 
 class Config(object):
     """ Global Tox config object. """
+
     def __init__(self, pluginmanager, option, interpreters):
         #: dictionary containing envname to envconfig mappings
         self.envconfigs = {}
@@ -601,6 +641,7 @@ class TestenvConfig:
     In addition to some core attributes/properties this config object holds all
     per-testenv ini attributes as attributes, see "tox --help-ini" for an overview.
     """
+
     def __init__(self, envname, config, factors, reader):
         #: test environment name
         self.envname = envname
@@ -619,9 +660,8 @@ class TestenvConfig:
 
     def get_envbindir(self):
         """ path to directory where scripts/binaries reside. """
-        if (sys.platform == "win32"
-                and "jython" not in self.basepython
-                and "pypy" not in self.basepython):
+        if sys.platform == "win32" and "jython" not in self.basepython and \
+                "pypy" not in self.basepython:
             return self.envdir.join("Scripts")
         else:
             return self.envdir.join("bin")
@@ -668,9 +708,6 @@ class TestenvConfig:
         if not info.version_info:
             raise tox.exception.InvocationError(
                 'Failed to get version_info for %s: %s' % (info.name, info.err))
-        if info.version_info < (2, 6):
-            raise tox.exception.UnsupportedInterpreter(
-                "python2.5 is not supported anymore, sorry")
         return info.executable
 
 
@@ -725,7 +762,7 @@ class parseini:
 
         reader.addsubstitutions(toxinidir=config.toxinidir,
                                 homedir=config.homedir)
-        # As older versions of tox may have bugs or incompatabilities that
+        # As older versions of tox may have bugs or incompatibilities that
         # prevent parsing of tox.ini this must be the first thing checked.
         config.minversion = reader.getstring("minversion", None)
         if config.minversion:
@@ -800,9 +837,8 @@ class parseini:
             if section in self._cfg or factors <= known_factors:
                 config.envconfigs[name] = self.make_envconfig(name, section, reader._subs, config)
 
-        all_develop = all(name in config.envconfigs
-                          and config.envconfigs[name].usedevelop
-                          for name in config.envlist)
+        all_develop = all(name in config.envconfigs and
+                          config.envconfigs[name].usedevelop for name in config.envlist)
 
         config.skipsdist = reader.getbool("skipsdist", all_develop)
 
@@ -810,8 +846,8 @@ class parseini:
         factors = set()
         if section in self._cfg:
             for _, value in self._cfg[section].items():
-                exprs = re.findall(r'^([\w{}\.,-]+)\:\s+', value, re.M)
-                factors.update(*mapcat(_split_factor_expr, exprs))
+                exprs = re.findall(r'^([\w{}\.!,-]+)\:\s+', value, re.M)
+                factors.update(*mapcat(_split_factor_expr_all, exprs))
         return factors
 
     def make_envconfig(self, name, section, subs, config, replace=True):
@@ -845,14 +881,14 @@ class parseini:
         return tc
 
     def _getenvdata(self, reader):
-        envstr = self.config.option.env                                \
-            or os.environ.get("TOXENV")                                \
-            or reader.getstring("envlist", replace=False) \
-            or []
+        envstr = self.config.option.env \
+                 or os.environ.get("TOXENV") \
+                 or reader.getstring("envlist", replace=False) \
+                 or []
         envlist = _split_env(envstr)
 
         # collect section envs
-        all_envs = set(envlist) - set(["ALL"])
+        all_envs = set(envlist) - {"ALL"}
         for section in self._cfg:
             if section.name.startswith(testenvprefix):
                 all_envs.add(section.name[len(testenvprefix):])
@@ -868,15 +904,37 @@ class parseini:
 def _split_env(env):
     """if handed a list, action="append" was used for -e """
     if not isinstance(env, list):
-        if '\n' in env:
-            env = ','.join(env.split('\n'))
+        env = [e.split('#', 1)[0].strip() for e in env.split('\n')]
+        env = ','.join([e for e in env if e])
         env = [env]
     return mapcat(_expand_envstr, env)
 
 
+def _is_negated_factor(factor):
+    return factor.startswith('!')
+
+
+def _base_factor_name(factor):
+    return factor[1:] if _is_negated_factor(factor) else factor
+
+
 def _split_factor_expr(expr):
+    def split_single(e):
+        raw = e.split('-')
+        included = {_base_factor_name(factor) for factor in raw
+                    if not _is_negated_factor(factor)}
+        excluded = {_base_factor_name(factor) for factor in raw
+                    if _is_negated_factor(factor)}
+        return included, excluded
+
     partial_envs = _expand_envstr(expr)
-    return [set(e.split('-')) for e in partial_envs]
+    return [split_single(e) for e in partial_envs]
+
+
+def _split_factor_expr_all(expr):
+    partial_envs = _expand_envstr(expr)
+    return [{_base_factor_name(factor) for factor in e.split('-')}
+            for e in partial_envs]
 
 
 def _expand_envstr(envstr):
@@ -887,7 +945,7 @@ def _expand_envstr(envstr):
 
     def expand(env):
         tokens = re.split(r'\{([^}]+)\}', env)
-        parts = [token.split(',') for token in tokens]
+        parts = [re.sub('\s+', '', token).split(',') for token in tokens]
         return [''.join(variant) for variant in itertools.product(*parts)]
 
     return mapcat(expand, envlist)
@@ -908,6 +966,7 @@ class DepConfig:
                 return self.name
             return ":%s:%s" % (self.indexserver.name, self.name)
         return str(self.name)
+
     __repr__ = __str__
 
 
@@ -919,7 +978,7 @@ class IndexServerConfig:
 
 #: Check value matches substitution form
 #: of referencing value from other section. E.g. {[base]commands}
-is_section_substitution = re.compile("{\[[^{}\s]+\]\S+?}").match
+is_section_substitution = re.compile(r"{\[[^{}\s]+\]\S+?}").match
 
 
 class SectionReader:
@@ -947,9 +1006,9 @@ class SectionReader:
             self.posargs = _posargs
 
     def getpath(self, name, defaultpath, replace=True):
-        toxinidir = self._subs['toxinidir']
         path = self.getstring(name, defaultpath, replace=replace)
         if path is not None:
+            toxinidir = self._subs['toxinidir']
             return toxinidir.join(path, abs=True)
 
     def getlist(self, name, sep="\n"):
@@ -986,7 +1045,7 @@ class SectionReader:
             s = default
         if s is None:
             raise KeyError("no config value [%s] %s found" % (
-                           self.section_name, name))
+                self.section_name, name))
 
         if not isinstance(s, bool):
             if s.lower() == "true":
@@ -1026,12 +1085,14 @@ class SectionReader:
 
     def _apply_factors(self, s):
         def factor_line(line):
-            m = re.search(r'^([\w{}\.,-]+)\:\s+(.+)', line)
+            m = re.search(r'^([\w{}\.!,-]+)\:\s+(.+)', line)
             if not m:
                 return line
 
             expr, line = m.groups()
-            if any(fs <= self.factors for fs in _split_factor_expr(expr)):
+            if any(included <= self.factors and not
+                   any(x in self.factors for x in excluded)
+                   for included, excluded in _split_factor_expr(expr)):
                 return line
 
         lines = s.strip().splitlines()
@@ -1073,6 +1134,7 @@ class Replacer:
         '''
         Recursively expand substitutions starting from the innermost expression
         '''
+
         def substitute_once(x):
             return self.RE_ITEM_REF.sub(self._replace_match, x)
 
@@ -1150,7 +1212,7 @@ class Replacer:
         val = self.reader._subs.get(sub_key, None)
         if val is None:
             val = self._substitute_from_other_section(sub_key)
-        if py.builtin.callable(val):
+        if callable(val):
             val = val()
         return str(val)
 
@@ -1230,7 +1292,6 @@ class _ArgvlistReader:
 
 
 class CommandParser(object):
-
     class State(object):
         def __init__(self):
             self.word = ''
